/*
 * $Id: terrace.c,v 1.5 2005/01/11 17:19:00 pdclose Exp $
 *
 * Copyright  2002, 2003, 2004, 2005 Paul Close
 * 
 * This software is released as OSI Certified Open Source under the
 * Artistic License (see LICENSE.txt) http://www.opensource.org/
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include "cards.h"

/*
 * Terrace
 *
 * Tableaus build down circularly in alternating colors.  Top cards may be
 * moved to the foundation, or among the tableaus.  Empty tableaus will be
 * filled automatically from the discard.
 *
 * Foundations build up circularly in alternating colors.  Empty
 * foundations must be filled with cards of the same rank as the
 * "starter" card that is chosen as the first move.
 *
 * Cards from the stock (the fan along the top) may only be played on the
 * foundation.
 *
 * Deal a card onto the discard pile by tapping the deck.  If emptied, the
 * discard will automatically deal a new card from the deck.
 *
 * Goal: move all cards to the foundations.
 *
 * Note: The game starts with only a few cards in the tableaus.  Examine the
 * stock carefully, then choose one of the tableau cards to be the "starter"
 * card by moving it to the foundation.  No other action is allowed until
 * this is done.  Once a "starter" card is chosen, the rest of the tableau
 * will be dealt, and normal game play will start.
 */

#define TerraceDeck	0
#define TerraceDiscard	1
#define TerracePile	2
#define TerraceNpile	8
#define TerraceLpile	(TerracePile + TerraceNpile - 1)
#define TerraceTable	(TerraceLpile + 1)
#define TerraceNtable	9
#define TerraceLtable	(TerraceTable + TerraceNtable - 1)
#define TerraceStock	(TerraceLtable + 1)
#define TerraceNstock	1
#define TerraceLstock	(TerraceStock + TerraceNstock - 1)
#define TerraceNstack	(TerraceLstock + 1)

CardSetupType	TerraceSetup[] =  {
    // deck
    { { 8, 0 }, { 0, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartDown },

    // discard
    { { 7, 0 }, { 1, 0 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },

    // pile / foundation
    { { 0, 1 }, { 1, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 1, 1 }, { 2, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 2, 1 }, { 3, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 3, 1 }, { 4, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 4, 1 }, { 5, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 5, 1 }, { 6, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 6, 1 }, { 7, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },
    { { 7, 1 }, { 8, 1 }, CardEmpty, CardDisplayTopCount, False, False, 0, CardStartUp },

    // table
    { { 0, 2 }, { 0, 2 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 1, 2 }, { 1, 2 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 2, 2 }, { 2, 2 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 3, 2 }, { 3, 2 }, CardNone, CardDisplayAll, False, False, 1, CardStartUp },
    { { 4, 2 }, { 4, 2 }, CardNone, CardDisplayAll, False, False, 0, CardStartUp },
    { { 5, 2 }, { 5, 2 }, CardNone, CardDisplayAll, False, False, 0, CardStartUp },
    { { 6, 2 }, { 6, 2 }, CardNone, CardDisplayAll, False, False, 0, CardStartUp },
    { { 7, 2 }, { 7, 2 }, CardNone, CardDisplayAll, False, False, 0, CardStartUp },
    { { 8, 2 }, { 8, 2 }, CardNone, CardDisplayAll, False, False, 0, CardStartUp },
    
    // stock
    { { 1, 0 }, { 3, 0 }, CardNone, CardDisplayAll, True, False, 11, CardStartUp },
};

static Card TerraceBase (void) GAME_SECT;
static CardProgressType TerraceCheckTable (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType TerraceCheckPile (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static CardProgressType TerraceCheckDeal (CardStackIndex src,
        CardStackIndex dst, CardIndex card) GAME_SECT;
static void TerraceMoveTable (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static void TerraceMoveDiscard (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static void TerraceMoveDeal (CardStackIndex src, CardStackIndex dst,
        CardIndex last) GAME_SECT;
static CardBool TerraceNormalize (void) GAME_SECT;

static Card
TerraceBase (void)
{
    CardStackType   *sp = CardStackPtr (TerracePile);
#define NOBASE 0

    if (sp->first != NullIndex)
	return Rank (CardPtr (sp->first)->card);
    return NOBASE;
}


static CardProgressType
TerraceCheckTable (CardStackIndex  src, 
		    CardStackIndex  dst,
		    CardIndex	    card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (!CardPtr (card)->face)
	return CardProgressCant;
    if (dp->last == NullIndex)
    {
        // Since we autofill from discard or deck, and you can't fill the
        // table from other sources, if there's an empty table, there's
        // nothing we can do
        return CardProgressCant;
    }
    else
    {
        CardType        *firstp, *lastp, *cardp;

        firstp = CardPtr(dp->first);
        lastp = CardPtr(dp->last);
        cardp = CardPtr(card);

        // down in alternating suit
        if (Rank(cardp->card) != Rank(firstp->card) &&
                CardIsInAlternatingSuitCircularOrder (card, dp->last))
        {
            // Don't allow stacking if we haven't chosen a base yet
            if (TerraceBase () == NOBASE)
                return CardProgressCant;
            // If more than one card in the table stack, by definition they
            // are in order, so don't move the top card somewhere else
            if (TerraceTable <= src && src <= TerraceLtable &&
                    cardp->prev != NullIndex)
		return CardProgressNone;
            return CardProgressSome;
        }
    }
    return CardProgressCant;
}

static CardProgressType
TerraceCheckPile (CardStackIndex src, 
		   CardStackIndex dst,
		   CardIndex	card)
{
    CardStackType   *dp = CardStackPtr (dst);

    if (!CardPtr (card)->face)
	return CardProgressCant;
    if (dp->last == NullIndex)
    {
        // if we haven't chosen a base yet, only allow a move from the table
        if (TerraceBase () == NOBASE)
        {
            if (TerraceTable <= src && src <= TerraceLtable)
                return CardProgressSome;
            return CardProgressCant;
        }
        // otherwise only allow a base rank to start this stack
        if (Rank (CardPtr (card)->card) == TerraceBase())
	    return CardProgressSome;
    }
    else
    {
        CardType        *firstp, *lastp, *cardp;

        firstp = CardPtr(dp->first);
        lastp = CardPtr(dp->last);
        cardp = CardPtr(card);

        // up in alternating suit
        if (Rank(cardp->card) != Rank(firstp->card) &&
                CardIsInAlternatingSuitCircularOrder (dp->last, card))
            return CardProgressSome;
    }
    return CardProgressCant;
}

static CardProgressType
TerraceCheckDeal (CardStackIndex src, 
		   CardStackIndex dst,
		   CardIndex	card)
{
    // don't allow deal until a base card has been chosen
    if (TerraceBase () == NOBASE)
        return CardProgressCant;
    return CardProgressSome;
}

static void
TerraceMoveTable (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    Boolean         first_card = false;

    // initial selection, make sure it goes into the first pile stack,
    // since that's what determines the base
    if (TerraceBase () == NOBASE)
    {
        dst = TerracePile;
        first_card = true;
    }

    CardMove (src, card, dst, True);

    if (first_card)
    {
        for (stack = TerracePile+1; stack <= TerraceLpile; stack++)
        {
            sp = CardStackPtr (stack);
            sp->empty = MakeCard(0, EMPTY, TerraceBase ());
            sp->dirty = true;
        }
    }

    // Autofill the table from the discard, or from the deck if the discard
    // is empty (normally only one stack, the src, will be empty, but after
    // the base card is chosen, we need to fill all empty spots; so always
    // fill all empty spots)

    //if (TerraceTable <= src && src <= TerraceLtable &&
    //        CardStackPtr (src)->last == NullIndex)
    for (stack = TerraceTable; stack <= TerraceLtable; stack++)
    {
        CardStackIndex  deck;
        CardIndex       deal;

	sp = CardStackPtr (stack);
	if (sp->last != NullIndex)
            continue;

        deck = TerraceDiscard;
        deal = CardStackPtr (deck)->last;

	if (deal == NullIndex)
        {
            deck = TerraceDeck;
            deal = CardStackPtr (deck)->last;
            if (deal == NullIndex)
                return;
        }
	CardMove (deck, deal, stack, True);
	CardTurn (src, sp->last, True, True);

        // autodeal the discard, if necessary
        if (CardStackPtr (TerraceDiscard)->last == NullIndex)
            TerraceMoveDeal (TerraceDeck, TerraceDiscard, 0);
    }
}

static void
TerraceMoveDiscard (CardStackIndex	src,
	       CardStackIndex	dst,
	       CardIndex	card)
{
    CardMove (src, card, dst, True);

    // autodeal the discard, if necessary
    if (CardStackPtr (TerraceDiscard)->last == NullIndex)
        TerraceMoveDeal (TerraceDeck, TerraceDiscard, 0);
}

static void
TerraceMoveDeal (CardStackIndex    src,
		  CardStackIndex    dst,
		  CardIndex	    last)
{
    CardIndex	card = CardStackPtr (src)->last;
    if (card != NullIndex)
    {
        CardMove (src, card, dst, True);
        CardTurn (dst, card, True, True);
    }
}

// stock to pile only
// table to pile or table
// discard to pile or table
CardMoveType TerraceMoves[] = {
    /* 0    Move card from stock to pile */
    {	"stock to pile",
	{ TerraceStock, TerraceStock },
	{ TerraceStock, TerraceStock },
	{ TerracePile, TerraceLpile },
	CardMoveAuto,
	0,
	TerraceCheckPile,
	0,
    },
    /* 1    Move card from table to pile */
    {	"table to pile",
	{ TerraceTable, TerraceLtable },
	{ TerraceTable, TerraceLtable },
	{ TerracePile, TerraceLpile },
	CardMoveAuto,
	0,
	TerraceCheckPile,
	TerraceMoveTable,
    },
    /* 2    Move card from discard to pile */
    {	"discard to pile",
	{ TerraceDiscard, TerraceDiscard },
	{ TerraceDiscard, TerraceDiscard },
	{ TerracePile, TerraceLpile },
	CardMoveAuto,
	0,
	TerraceCheckPile,
	TerraceMoveDiscard,
    },
    /* 3    Move card from discard to table */
    {	"discard to table",
	{ TerraceDiscard, TerraceDiscard },
	{ TerraceDiscard, TerraceDiscard },
	{ TerraceTable, TerraceLtable },
	CardMoveHelpful,
	0,
	TerraceCheckTable,
	TerraceMoveDiscard,
    },
    /* 4    Move card from table to table */
    {	"table to table",
	{ TerraceTable, TerraceLtable },
	{ TerraceTable, TerraceLtable },
	{ TerraceTable, TerraceLtable },
	CardMoveHelpful,
	0,
	TerraceCheckTable,
	TerraceMoveTable,
    },
    /* 5    Deal next card */
    {	"deal",
	{ TerraceDeck, TerraceDeck },
	{ TerraceDeck, TerraceDeck },
	{ TerraceDiscard, TerraceDiscard },
	CardMoveHelpful,
	0,
	TerraceCheckDeal,
	TerraceMoveDeal,
    },
};


#define NMOVES	(sizeof (TerraceMoves) / sizeof (TerraceMoves[0]))

static CardBool
TerraceNormalize (void)
{
    CardStackIndex  stack;
    CardStackType   *sp;
    CardIndex	    card;
    CardType	    *cp;
    Card            target;

    target = TerraceBase () - 1;
    if (target == 0)
        target = KING;

    for (stack = TerracePile; stack <= TerraceLpile; stack++)
    {
	sp = CardStackPtr (stack);
	card = sp->last;
	if (card == NullIndex)
	    return False;
	cp = CardPtr (card);
	if (Rank (cp->card) != target)
	    return False;
    }
    return True;
}

RulesType   terraceGame = {
    MakeCard (0, CLUB, 1), MakeCard (0, SPADE, KING), 2, TerraceNstack,
    TerraceNormalize,
    0,
    0,
    TerraceSetup,
    NMOVES,
    TerraceMoves
};
